program countdown
    use mpi
    implicit none
    ! Timer still ticking, messages contains the timer value
    integer, parameter :: TAG_TICKER = 1
    ! Bomb has exploded, message contains the rank of the loser
    integer, parameter :: TAG_KABOOM = 2
    integer :: ierr, myRank, numProcs
    integer :: nextRank, loserRank
    integer :: countDownCounter
    logical :: done
    integer, dimension(MPI_STATUS_SIZE) :: status
    integer :: request
    integer :: data, i
    character(len=MPI_MAX_PROCESSOR_NAME) :: myString
    integer :: myStringLength

    integer, external :: random_dec

    ! MPI Initialization
    call MPI_INIT(ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, numProcs, ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, myRank, ierr)

    ! Rank-specific initialization of the random number generator
    call srand(int(MPI_Wtime()) + myRank*100)

    nextRank = mod(myRank+1, numProcs)
    loserRank = -1

    countDownCounter = 50

    ! Rank 0 starts the bomb
    if (myRank == 0) then
        if (iargc() > 0) then
            call getarg(1, myString)
            read(myString,*) countDownCounter
            if (countDownCounter <= 0) stop
        end if
        print '("Counting down from ",I3)', countDownCounter
        call MPI_SEND(countDownCounter, 1, MPI_INTEGER, nextRank, TAG_TICKER, &
            MPI_COMM_WORLD, ierr)
    end if

    call MPI_GET_PROCESSOR_NAME(myString, myStringLength, ierr)
    print '("I am rank ",I3," running on ",A)', myRank, trim(myString)

    done = .false.
    request = MPI_REQUEST_NULL

    do while (.not.done)
        call MPI_RECV(data, 1, MPI_INTEGER, MPI_ANY_SOURCE, MPI_ANY_TAG, &
            MPI_COMM_WORLD, status, ierr)
        if (status(MPI_TAG) == TAG_KABOOM) then
            ! The shutdown message was received
            done = .true.
            loserRank = data
        else
            countDownCounter = random_dec(data)
            if (countDownCounter > 0) then
                print '("Process ",I3," has received the bomb (",I4," on the clock)&
                    & and is still alive!")', myRank, countDownCounter
                call MPI_SEND(countDownCounter, 1, MPI_INTEGER, nextRank, TAG_TICKER, &
                    MPI_COMM_WORLD, ierr)
            else
                print '("Process ",I3," lost")', myRank
                do i = 0, numProcs-1
                    if (i /= myRank) then
                        call MPI_SEND(myRank, 1, MPI_INTEGER, i, TAG_KABOOM, &
                            MPI_COMM_WORLD, ierr)
                    else
                        ! Make sure that send to self will not block
                        call MPI_ISEND(myRank, 1, MPI_INTEGER, i, TAG_KABOOM, &
                            MPI_COMM_WORLD, request, ierr)
                    end if
                end do
            end if
        end if
    end do
    if (request /= MPI_REQUEST_NULL) call MPI_WAIT(request, status, ierr)

    print '("I am process ",I3," and ",I3," is the loser")', myRank, loserRank

    call MPI_FINALIZE(ierr)
end program countdown


function random_dec(current)
    ! Comment out the following USE statement when using GCC
    use ifport
    implicit none
    integer, intent(in) :: current
    integer :: random_dec

    random_dec = current - mod(irand(), 10)
    if (random_dec < 0) random_dec = 0
end function random_dec
